using System.Security.Permissions;
using System;
using System.Threading;
using System.Collections.Concurrent;
using WingServerCommon.Utilities;
using WingServerCommon.Log;
using WingInterfaceLibrary.Request.Authentication;
using WingInterfaceLibrary.Result;
using WingInterfaceLibrary.Enum;
using System.Collections.Generic;
using System.Linq;
using WingServerCommon.Config;
using WingServerCommon.Config.Parameters;

using System.Threading.Tasks;
using WingInterfaceLibrary.Interface;
using WingServerCommon.Interfaces.Cache;
using WingServerCommon.Mapper;
using System.Net;
using WingInterfaceLibrary.DTO.Management;
using WingInterfaceLibrary.Enum;
using System.IO;
using System.Text.RegularExpressions;

namespace WingSessionModule
{
    interface ITokensManagerExt : ITokensManager
    {
        /// <summary>
        /// Get paged tokens 
        /// </summary>
        /// <param name="keyword">The keuword filter</param>
        /// <param name="pageIndex">The page index</param>
        /// <param name="pageSize">The page size</param>
        /// <param name="totalCount">The total token count</param>
        /// <returns></returns>
        IList<Token> GetPagedTokens(string keyword, int pageIndex, int pageSize, int accountType, int loginSource, string isOnline, out int totalCount);

        /// <summary>
        /// Add a new token
        /// </summary>
        /// <param name="accountType">Account Type</param>
        /// <param name="accountName">Account name</param>
        /// <param name="clientId">Client id</param>
        /// <param name="platform">Login Platform</param>
        /// <param name="loginSource">Login srouce</param>
        /// <param name="loginServer">Login server</param>
        /// <param name="clientIp">The client ip</param>
        /// <returns></returns>
        Task<Token> AddAsync(AccountType accountType, string accountName, string clientId, Platform platform, LoginSource loginSource, string loginServer, long clientIp, string installVersion, bool isOldPlatform, ProxyTypeEnum proxyType);

        /// <summary>
        /// 重新加载用户黑名单
        /// </summary>
        void RefreshBlackClientIds();

        /// <summary>
        /// Validate the token string
        /// </summary>
        /// <param name="tokenString"></param>
        /// <returns></returns>
        ValidateTokenResult Validate(string tokenString, string methodName);

        /// <summary>
        /// Get tokens with specified client id
        /// </summary>
        /// <param name="clientId">User code, admin code or device code</param>
        /// <returns></returns>
        IList<Token> GetTokens(string clientId);

        /// <summary>
        /// get tokenDTO by token
        /// </summary>
        /// <param name="token"></param>
        /// <returns></returns>
        TokenDTO GetTokenDTO(string token);

        /// <summary>
        /// 移除延长token
        /// </summary>
        /// <param name="tokenCode">token Code</param>
        /// <returns></returns>
        bool RemoveExtendToken(string tokenCode);

        /// <summary>
        /// Get tokens with specified client ids
        /// </summary>
        /// <param name="tokenValues">client id array</param>
        /// <returns></returns>
        IList<Token> GetTokenWithClientIds(List<string> clientIds);

        /// <summary>
        /// Get tokens with specified token values
        /// </summary>
        /// <param name="tokenValues">token value array</param>
        /// <returns></returns>
        IList<Token> GetTokenWithValues(List<string> tokenValues);

        /// <summary>
        /// The token replace event
        /// </summary>
        event EventHandler<TokenEventArgs> TokenReplaced;
    }


    internal class ServerTokenManager : CacheManager<Token>, ITokensManagerExt
    {
        static int TokenVersion = 1;

        private int Expiration = 52560000 / 2;

        private ConcurrentDictionary<string, Token> _extendTokens = new ConcurrentDictionary<string, Token>();//延长token

        private bool _isWatching = true;
        private TokenValidator _tokenValidator;
        public event EventHandler<TokenEventArgs> TokenReplaced;
        public Func<IList<Token>, Task<bool>> TokenUpdate;
        public Func<TokenAddRequest, Task<bool>> TokenAdd;
        public Func<Task<bool>> RemoveExpiredToken;

        public Func<string, Task<ThirdPartyInfoDTO>> GetThirdPartyInfo;

        public ServerTokenManager(Func<string, Token> loadCacheFromDbWithId, Func<IList<string>, IList<Token>> loadCachesFromDbWithIds, Func<IList<Token>, Task<bool>> tokenUpdate, Func<TokenAddRequest, Task<bool>> tokenAdd, Func<Task<bool>> removeExpiredToken, Func<string, Task<ThirdPartyInfoDTO>> getThirdPartyInfo) : base(loadCacheFromDbWithId, loadCachesFromDbWithIds)
        {
            CacheMaintenance.Instance.Get<ITokensManager>().RegisterDBFunction(loadCacheFromDbWithId, loadCachesFromDbWithIds);
            var rules = new List<TokenValidatorRule>() { new TokenValidatorExpireRule(), new TokenValidatorVersionRule(), new TokenBlackListRule(), new ThirdPartyInterfaceRule(getThirdPartyInfo) };
            _tokenValidator = new TokenValidator(rules);
            TokenUpdate = tokenUpdate;
            TokenAdd = tokenAdd;
            RemoveExpiredToken = removeExpiredToken;
            GetThirdPartyInfo = getThirdPartyInfo;
            WatchingTokenTask();
        }

        /// <summary>
        /// 重新加载用户黑名单
        /// </summary>
        public void RefreshBlackClientIds()
        {
            _tokenValidator.RefreshBlackListRule();
        }

        /// <summary>
        /// Validate the token string
        /// </summary>
        /// <param name="tokenString"></param>
        /// <returns></returns>
        public ValidateTokenResult Validate(string tokenString, string methodName)
        {
            var code = CustomerRpcCode.Unknow;
            Token token = null;
            try
            {
                token = CacheMaintenance.Instance.Get<ITokensManager>().Get(tokenString);
            }
            catch (InvalidOperationException ex)
            {
                code = CustomerRpcCode.TokenNotExist;
            }
            if (token == null)
            {
                code = CustomerRpcCode.TokenNotExist;
            }
            else
            {
                code = _tokenValidator.Execute(token, methodName);
                if (code == CustomerRpcCode.Ok)
                {
                    //activate token
                    token.Expiration = DateTime.UtcNow.AddMinutes(GetTokenDuration(token.AccountType));
                    //延长token监控
                    if (token.Expiration < DateTime.UtcNow.AddMinutes(Expiration))
                    {
                        var extendToken = GetExtendToken(token.Code);
                        if (extendToken == null)
                        {
                            _extendTokens.TryAdd(token.Code, token);
                        }
                        else
                        {
                            extendToken.Expiration = token.Expiration;
                        }
                    }
                }
            }
            var tokenDTO = token.MappingTo<TokenDTO>();
            return new ValidateTokenResult(code, tokenDTO);
        }

        /// <summary>
        /// Get token with specified the token string
        /// </summary>
        /// <param name="tokenString"></param>
        /// <returns></returns>
        private Token GetExtendToken(string tokenString)
        {
            _extendTokens.TryGetValue(tokenString, out var token);
            return token;
        }

        /// <summary>
        /// Add a new token
        /// </summary>
        /// <param name="token"></param>
        public async Task<Token> AddAsync(AccountType accountType, string accountName, string clientId, Platform platform, LoginSource loginSource, string loginServer, long clientIp, string installVersion, bool isOldPlatform, ProxyTypeEnum proxyType)
        {
            var tokens = CacheMaintenance.Instance.Get<ITokensManager>().Where(x => x.ClientId == clientId)?.ToList() ?? new List<Token>();
            if (tokens.Count > 1)
            {
                //TODO should exception, should clean these tokens ?
                //一个client是可以有多个token的
            }

            var value = IdHelper.Generate<string>();
            var timeoutMinutes = GetTokenDuration((int)accountType);
            //token timeout should be set in app settings
            var newToken = new Token
            {
                Version = TokenVersion,
                Code = value,
                AccountType = (int)accountType,
                Platform = (int)platform,
                LoginSource = (int)loginSource,
                ClientId = clientId,
                LoginServer = loginServer,
                CreateTime = DateTime.UtcNow,
                Expiration = DateTime.UtcNow.AddMinutes(timeoutMinutes),
                IpValue = clientIp,
                AccountName = accountName,
                InstallVersion = installVersion,
                IsOnline = true,//连上后默认在线
                IsOldPlatform = isOldPlatform,
                ProxyType = (int)proxyType
            };
            var tokenAddRequest = new TokenAddRequest();
            tokenAddRequest.Token = newToken;
            tokenAddRequest.ClientId = clientId;
            tokenAddRequest.LoginSource = loginSource;
            await TokenAdd(tokenAddRequest);
            //第三方token处理
            if (accountType == AccountType.ThirdParty)
            {
                var existToken = tokens.FirstOrDefault();
                if (existToken != null)
                {
                    //Replace exist token    
                    OnTokenReplaced(existToken, newToken);
                }
                else
                {
                    CacheMaintenance.Instance.Get<ITokensManager>().Add(newToken);
                    Logger.WriteLineInfo($"Add new {newToken.ToString()}");
                }
            }
            else
            {
                var existToken = tokens.FirstOrDefault(t => GetLoginSource(loginSource).Contains(t.LoginSource));
                if (existToken != null)
                {
                    OnTokenReplaced(existToken, newToken);
                }
                else
                {
                    CacheMaintenance.Instance.Get<ITokensManager>().Add(newToken);
                    Logger.WriteLineInfo($"Add new {newToken.ToString()}");
                }
            }
            return newToken;
        }

        private List<int> GetLoginSource(LoginSource loginSource)
        {
            var list = new List<int>();
            switch (loginSource)
            {
                case LoginSource.PC:
                case LoginSource.Web:
                    list.Add((int)LoginSource.PC);
                    list.Add((int)LoginSource.Web);
                    break;
                case LoginSource.Mobile:
                case LoginSource.Pad:
                    list.Add((int)LoginSource.Mobile);
                    list.Add((int)LoginSource.Pad);
                    break;
                default:
                    list.Add((int)loginSource);
                    break;
            }
            return list;
        }

        private int GetTokenDuration(int accountType)
        {
            int tokenDuration = 20;
            switch ((AccountType)accountType)
            {
                case AccountType.Admin:
                    tokenDuration = ConfigurationManager.GetParammeter<IntParameter>("Authentication", "Admin").Value;
                    break;
                case AccountType.User:
                    tokenDuration = ConfigurationManager.GetParammeter<IntParameter>("Authentication", "User").Value;
                    break;
                case AccountType.US:
                    tokenDuration = ConfigurationManager.GetParammeter<IntParameter>("Authentication", "US").Value;
                    break;
                case AccountType.USBox:
                    tokenDuration = ConfigurationManager.GetParammeter<IntParameter>("Authentication", "USBox").Value;
                    break;
                case AccountType.ThirdParty:
                    tokenDuration = ConfigurationManager.GetParammeter<IntParameter>("Authentication", "ThirdParty").Value;
                    break;
            }
            return tokenDuration;
        }

        private void OnTokenReplaced(Token oldToken, Token newToken)
        {
            //发送给旧的 有验证 要在移除之前做
            if (newToken.AccountType == (int)AccountType.User)
            {
                TokenReplaced?.Invoke(this, new TokenEventArgs(oldToken, newToken));
            }
            //将遗留的老token全部移除
            var loginSources = GetLoginSource((LoginSource)newToken.LoginSource);
            var oldTokens = CacheMaintenance.Instance.Get<ITokensManager>().Where(x => x.ClientId == newToken.ClientId);
            oldTokens = oldTokens.Where(x => loginSources.Contains(x.LoginSource))?.ToList();
            foreach (var item in oldTokens)
            {
                CacheMaintenance.Instance.Get<ITokensManager>().Remove(item.Code);
                //延长的也删除，重新登录已经更新DB了，不需要延长了。
                _extendTokens.TryRemove(item.Code, out var extendRemoved);
            }
            CacheMaintenance.Instance.Get<ITokensManager>().Add(newToken);
            Logger.WriteLineInfo($"Replaced token old {oldToken.ToString()} to new {newToken.ToString()}");
        }

        /// <summary>
        /// Get tokens with specified client id
        /// </summary>
        /// <param name="clientId">User code, admin code or device code</param>
        /// <returns></returns>
        public IList<Token> GetTokens(string clientId)
        {
            var tokens = CacheMaintenance.Instance.Get<ITokensManager>().Where(x => x.ClientId == clientId)?.ToList() ?? new List<Token>();
            if (tokens.Count > 0)
            {
                //如果出现单个设备多个token的情况，筛选出正确的那个token
                var rightTokens = new List<Token>();
                var deviceTokens = tokens.Where(x => x.AccountType == (int)AccountType.US || x.AccountType == (int)AccountType.USBox);
                var deviceCodes = deviceTokens.Select(x => x.ClientId);
                foreach (var deviceCode in deviceCodes)
                {
                    var deviceToken = tokens.Where(x => x.ClientId == deviceCode).OrderByDescending(x => x.CreateTime).FirstOrDefault();
                    rightTokens.Add(deviceToken);
                }
                tokens = tokens.Where(x => !deviceCodes.Contains(x.ClientId)).ToList();
                tokens.AddRange(rightTokens);
            }
            return tokens;
        }

        /// <summary>
        /// 移除延长token
        /// </summary>
        /// <param name="tokenCode">token Code</param>
        /// <returns></returns>
        public bool RemoveExtendToken(string tokenCode)
        {
            var result = true;
            if (_extendTokens.ContainsKey(tokenCode))
            {
                result = _extendTokens.TryRemove(tokenCode, out _);
            }
            return result;
        }


        /// <summary>
        /// Get tokenDTO by token
        /// </summary>
        /// <param name="clientId">User code, admin code or device code</param>
        /// <returns></returns>
        public TokenDTO GetTokenDTO(string token)
        {
            var tokenDTO = new TokenDTO();
            var tokenList = CacheMaintenance.Instance.Get<ITokensManager>().Where(c => c.Code == token);
            if (tokenList?.Count > 0)
            {
                Token tokenCache = tokenList.FirstOrDefault();
                tokenDTO = tokenCache.MappingTo<TokenDTO>();
            }
            return tokenDTO;
        }

        /// <summary>
        /// Get tokens with specified client ids
        /// </summary>
        /// <param name="tokenValues">client id array</param>
        /// <returns></returns>
        public IList<Token> GetTokenWithClientIds(List<string> clientIds)
        {
            var tokens = CacheMaintenance.Instance.Get<ITokensManager>().Where(x => clientIds.Contains(x.ClientId))?.ToList() ?? new List<Token>();
            if (tokens.Count > 0)
            {
                //如果出现单个设备多个token的情况，筛选出正确的那个token
                var rightTokens = new List<Token>();
                var deviceTokens = tokens.Where(x => x.AccountType == (int)AccountType.US || x.AccountType == (int)AccountType.USBox);
                var deviceCodes = deviceTokens.Select(x => x.ClientId);
                foreach (var deviceCode in deviceCodes)
                {
                    var deviceToken = tokens.Where(x => x.ClientId == deviceCode).OrderByDescending(x => x.CreateTime).FirstOrDefault();
                    rightTokens.Add(deviceToken);
                }
                tokens = tokens.Where(x => !deviceCodes.Contains(x.ClientId)).ToList();
                tokens.AddRange(rightTokens);
            }
            return tokens;
        }

        /// <summary>
        /// Get tokens with specified token values
        /// </summary>
        /// <param name="tokenValues">token value array</param>
        /// <returns></returns>
        public IList<Token> GetTokenWithValues(List<string> tokenValues)
        {
            return CacheMaintenance.Instance.Get<ITokensManager>().Where(x => tokenValues.Contains(x.Code))?.ToList() ?? new List<Token>();
        }

        /// <summary>
        ///  watching tokenqueue need ync to DB
        /// </summary>
        private void WatchingTokenTask()
        {
            Task.Run(async () =>
            {
                try
                {
                    while (_isWatching)
                    {
                        await Task.Delay(TimeSpan.FromSeconds(ConfigurationManager.GetParammeter<IntParameter>("Authentication", "IntervalSec").Value));
                        //await Task.Delay(TimeSpan.FromSeconds(20));
                        if (_extendTokens.Count > 0)
                        {
                            Logger.WriteLineInfo($"Begin Watching tokenqueue data");

                            var tokens = _extendTokens.Select(t => t.Value).ToList();

                            await TokenUpdate(tokens);

                            Logger.WriteLineInfo($"end Watching tokenqueue data  count:{_extendTokens.Count}");
                            _extendTokens = new ConcurrentDictionary<string, Token>();
                        }
                        //过期token移除
                        await RemoveExpiredToken();

                    }
                }
                catch (Exception e)
                {
                    Logger.WriteLineError($"AuthenticationService WatchingTokenQueueTask error:{e}");
                }
            });
        }

        /// <summary>
        /// Get paged tokens 
        /// </summary>
        /// <param name="keyword">The keuword filter</param>
        /// <param name="pageIndex">The page index</param>
        /// <param name="pageSize">The page size</param>
        /// <param name="totalCount">The total token count</param>
        /// <returns></returns>
        public IList<Token> GetPagedTokens(string keyword, int pageIndex, int pageSize, int accountType, int loginSource, string isOnline, out int totalCount)
        {
            var lowerKeyword = string.IsNullOrEmpty(keyword) ? string.Empty : keyword.ToLower();
            IList<Token> result = null;
            var filterList = CacheMaintenance.Instance.Get<ITokensManager>().Where(x => x.Code.ToLower().Contains(lowerKeyword) || x.AccountName.ToLower().Contains(lowerKeyword) || IsExistIP(x.IpValue, lowerKeyword));
            if (accountType > -1)
            {
                filterList = filterList.Where(x => x.AccountType == accountType)?.ToList();
            }
            if (loginSource > -1)
            {
                filterList = filterList.Where(x => x.LoginSource == loginSource)?.ToList();
            }
            if (!string.IsNullOrEmpty(isOnline))
            {
                filterList = filterList.Where(x => x.IsOnline == Convert.ToBoolean(isOnline))?.ToList();
            }
            totalCount = filterList.Count;
            result = filterList.OrderByDescending(x => x.CreateTime).Skip((pageIndex - 1) * pageSize).Take(pageSize).ToList();
            return result;
        }

        /// <summary>
        /// 是否存在IP
        /// </summary>
        private bool IsExistIP(long ipValue, string keyword)
        {
            var ipAddress = IpLongToString(ipValue);
            if (ipAddress.Contains(keyword))
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// IP long转string
        /// </summary>
        private string IpLongToString(long ipValue)
        {
            if (ipValue > 0)
            {
                uint netInt = (uint)IPAddress.HostToNetworkOrder((Int32)ipValue);
                IPAddress ipaddr = new IPAddress((long)netInt);
                var ipAddress = ipaddr.ToString();
                return ipAddress;
            }
            else
            {
                return "";
            }
        }

        /// <summary>
        /// 清空缓存
        /// </summary>
        public void Dispose()
        {
            _extendTokens.Clear();
        }

    }

    internal class TokenValidator
    {
        readonly IList<TokenValidatorRule> _rules;
        public TokenValidator(IList<TokenValidatorRule> rules)
        {
            _rules = rules;
        }

        public CustomerRpcCode Execute(Token token, string methodName)
        {
            foreach (var rule in _rules)
            {
                var result = rule.Execute(token, methodName);
                if (result != CustomerRpcCode.Ok)
                {
                    return result;
                }
            }

            return CustomerRpcCode.Ok;
        }

        /// <summary>
        /// 刷新用户黑名单
        /// </summary>
        public void RefreshBlackListRule()
        {
            foreach (var rule in _rules)
            {
                if (rule is TokenBlackListRule blackListRule)
                {
                    blackListRule.LoadData();
                }
            }
        }
    }

    internal abstract class TokenValidatorRule
    {
        public abstract CustomerRpcCode Execute(Token token, string methodName);
    }

    internal class TokenValidatorExpireRule : TokenValidatorRule
    {
        public override CustomerRpcCode Execute(Token token, string methodName)
        {
            var now = DateTime.UtcNow;
            if (now > token.Expiration)
            {
                return CustomerRpcCode.TokenExpired;
            }

            return CustomerRpcCode.Ok;
        }
    }

    internal class TokenValidatorVersionRule : TokenValidatorRule
    {
        public override CustomerRpcCode Execute(Token token, string methodName)
        {
            var version = 0; //TODO load config and design logic
            if (token.Version < version)
            {
                return CustomerRpcCode.InvalidTokenVersion;
            }

            return CustomerRpcCode.Ok;
        }
    }

    internal class TokenBlackListRule : TokenValidatorRule
    {
        private string _folderPath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "Resource", "BlackClients");
        private HashSet<string> _blackClientIds;
        public TokenBlackListRule()
        {
            _blackClientIds = new HashSet<string>();
            if (!Directory.Exists(_folderPath))
            {
                Directory.CreateDirectory(_folderPath);
            }
            LoadData();
        }

        public override CustomerRpcCode Execute(Token token, string methodName)
        {
            if (_blackClientIds.Contains(token.ClientId))
            {
                return CustomerRpcCode.IPInBlacklist;
            }

            return CustomerRpcCode.Ok;
        }

        public void LoadData()
        {
            Logger.WriteLineInfo($"AuthenticationService TokenBlackListRule LoadData start");
            var fileInfos = Directory.GetFiles(_folderPath);
            if (fileInfos?.Any() ?? false)
            {
                var blackClientIds = Directory.GetFiles(_folderPath)
                .SelectMany(GetClientIds)
                .ToHashSet();

                _blackClientIds = blackClientIds;

                var blackClientIdCount = _blackClientIds.Count;
                Logger.WriteLineInfo($"AuthenticationService TokenBlackListRule LoadData success, black client count:{blackClientIdCount}");
                if (blackClientIdCount <= 10)
                {
                    Logger.WriteLineInfo($"AuthenticationService TokenBlackListRule LoadData success, black client ids:{string.Join(";", _blackClientIds)}");
                }
            }
            else
            {
                _blackClientIds = new HashSet<string>();
                Logger.WriteLineInfo($"AuthenticationService TokenBlackListRule LoadData success, folder is empty");
            }
        }

        private IEnumerable<string> GetClientIds(string filePath)
        {
            try
            {
                return File.ReadAllLines(filePath)
                    .Where(line => !string.IsNullOrWhiteSpace(line))
                    .SelectMany(line => Regex.Split(line, @"[^a-zA-Z0-9]"));
            }
            catch (Exception ex)
            {
                Logger.WriteLineInfo($"AuthenticationService TokenBlackListRule GetClientIds failed, filePath:{filePath}, ex:{ex}");
                return new List<string>();
            }
        }
    }

    internal class ThirdPartyInterfaceRule : TokenValidatorRule
    {
        public Func<string, Task<ThirdPartyInfoDTO>> GetThirdPartyInfo;
        public ThirdPartyInterfaceRule(Func<string, Task<ThirdPartyInfoDTO>> getThirdPartyInfo)
        {
            GetThirdPartyInfo = getThirdPartyInfo;
        }
        public override CustomerRpcCode Execute(Token token, string methodName)
        {
            if (token.AccountType == (int)AccountType.ThirdParty)
            {
                var info = GetThirdPartyInfo(token.Code).Result;
                if (!info.AllowedInterfaces.Contains(methodName))
                {
                    return CustomerRpcCode.AuthenticationError;
                }
            }
            return CustomerRpcCode.Ok;
        }
    }

    internal class TokenAddRequest
    {
        public Token Token { get; set; }

        public string ClientId { get; set; }

        public LoginSource LoginSource { get; set; }
    }
}