﻿using WingTokenModule.Common;
using System.Threading.Tasks;
using JsonRpcLite.Services;
using System;
using System.Collections.Generic;
using WingServerCommon.Service;
using WingInterfaceLibrary.Interface;
using WingInterfaceLibrary.Request;
using WingInterfaceLibrary.Result;
using WingInterfaceLibrary.Request.Authentication;
using WingSessionModule;
using WingServerCommon.Log;
using Newtonsoft.Json;
using WingInterfaceLibrary.Interface.DBInterface;
using WingInterfaceLibrary.DB.Request;
using WingInterfaceLibrary.Enum;
using WingServerCommon.Config;
using WingServerCommon.Config.Parameters;
using WingServerCommon.Mapper;
using WingServerCommon.Interfaces.Cache;
using System.Linq;
using System.Net;
using WingInterfaceLibrary.Notifications;
using WingInterfaceLibrary.Internal.Interface;
using WingInterfaceLibrary.Internal.Request;
using WingInterfaceLibrary.DTO.Management;

namespace WingTokenModule.Service
{
    /// <summary>
    /// 权限验证服务
    /// </summary>
    public class AuthenticationService : JsonRpcService, IAuthenticationService
    {
        private string DefaultToken = string.Empty;
        private readonly ITokensManagerExt _serverTokenManager;

        private INotificationService _notificationService;

        private ITokenDBService _tokenDBService;

        private IThirdPartyInfoDBService _thirdPartyInfoDBService;

        //是否体检系统本地工作站
        private bool _isLocalHost => GetStringParams("Vital", "CloudServerUrl") != "";

        public AuthenticationService()
        {
            _serverTokenManager = new ServerTokenManager(LoadTokenFormDbWithId, LoadTokensFormDbWithIds, OnTokenUpdate, OnTokenAdd, RemoveExpiredToken, GetThirdPartyInfo);
            _serverTokenManager.TokenReplaced += OnTokenReplaced;
        }

        /// <summary>
        /// /// Init service
        /// </summary>
        public override void Load(JsonRpcClientPool jsonRpcClientPool)
        {
            base.Load(jsonRpcClientPool);
            _tokenDBService = GetProxy<ITokenDBService>();
            _notificationService = GetProxy<INotificationService>();
            _thirdPartyInfoDBService = GetProxy<IThirdPartyInfoDBService>();
            DefaultToken = ConfigurationManager.GetParammeter<StringParameter>("Authentication", "BaseToken").Value;
        }

        public override async void LoadAfterRegister()
        {
            try
            {
                var tokenDtos = _tokenDBService.FindNotExpiredTokensAsync().Result;
                if (_isLocalHost)
                {
                    var tokens = new List<Token>();
                    foreach (var dto in tokenDtos)
                    {
                        //arm64mapping速度非常慢 超过20秒 原因暂时未知
                        var token = new Token();
                        token.AccountName = dto.AccountName;
                        token.AccountType = (int)dto.AccountType;
                        token.ClientId = dto.ClientId;
                        token.Code = dto.Code;
                        token.CreateTime = dto.CreateTime;
                        token.Expiration = dto.Expiration;
                        token.InstallVersion = dto.InstallVersion;
                        token.IpValue = dto.IpValue;
                        token.IsOldPlatform = dto.IsOldPlatform;
                        token.IsOnline = dto.IsOnline;
                        token.LoginServer = dto.LoginServer;
                        token.LoginSource = (int)dto.LoginSource;
                        token.Platform = (int)dto.Platform;
                        token.ProxyType = (int)dto.ProxyType;
                        token.Version = dto.Version;
                        tokens.Add(token);
                    }
                    CacheMaintenance.Instance.Get<ITokensManager>().LoadFromDbOject(tokens);
                }
                else
                {
                    var tokens = tokenDtos.MappingTo<IList<Token>>();
                    CacheMaintenance.Instance.Get<ITokensManager>().LoadFromDbOject(tokens);
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"AuthenticationService LoadDBToken err, {ex}");
            }
        }


        /// <summary>
        /// 通用异常处理方法
        /// </summary>
        /// <param name="customerRpcCodeEnum">异常标识码</param>
        /// <param name="msg">异常信息</param>
        /// <param name="internalMessage">内部异常信息，非必填</param>
        private void ThrowCustomerException(CustomerRpcCode customerRpcCodeEnum, string msg, string internalMessage = null)
        {
            ThrowRpcException((int)customerRpcCodeEnum, msg, internalMessage);
        }

        Token LoadTokenFormDbWithId(string id)
        {
            var tokenDB = _tokenDBService.FindTokenByValueAsync(id).Result;
            if (tokenDB != null)
            {
                return tokenDB.MappingTo<Token>();
            }
            return null;
        }

        IList<Token> LoadTokensFormDbWithIds(IList<string> ids)
        {
            var tokensDB = _tokenDBService.FindTokensWithIdsAsync(ids).Result;
            if (tokensDB != null)
            {
                return tokensDB.MappingTo<IList<Token>>();
            }
            return null;
        }

        private async Task<ThirdPartyInfoDTO> GetThirdPartyInfo(string token)
        {
            var result = await _thirdPartyInfoDBService.FindThirdPartyInfoByTokenAsync(token);
            return result;
        }

        /// <summary>
        /// 获取默认Token，项目内部调用
        /// </summary>
        public async Task<string> GetServerDefaultTokenAsync()
        {
            return DefaultToken;
        }

        /// <summary>
        /// 登出通知
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private void OnTokenReplaced(object sender, TokenEventArgs args)
        {
            var oldToken = args.OldItem;
            var oldTokens = new List<TokenDTO> { oldToken.MappingTo<TokenDTO>() };
            var tokenReplacedNotification = new TokenReplacedNotification
            {
                UserCode = oldToken.ClientId
            };
            Logger.WriteLineInfo($"Notify TokenReplacedNotification  user: {oldToken.ClientId},token:{oldToken.Code}");
            var notificationRequest = new SendNotificationRequest
            {
                ClientId = oldToken.ClientId,
                Message = tokenReplacedNotification,
                JsonMessage = JsonConvert.SerializeObject(tokenReplacedNotification),
                NotificationType = tokenReplacedNotification.NotificationType,
                TransactionType = TransactionTypeEnum.Session,
                RelevanceCode = args.OldItem.ClientId,
                ReceiverType = ApplicantTypeEnum.Client,
                Tokens = oldTokens
            };
            var result = _notificationService.PostMessageAsync(notificationRequest);
        }

        private async Task<bool> OnTokenUpdate(IList<Token> tokens)
        {
            foreach (var token in tokens)
            {
                try
                {
                    var tokenDB = await _tokenDBService.FindTokenByValueAsync(token.Code);
                    if (tokenDB != null && !string.IsNullOrWhiteSpace(tokenDB.Code))
                    {
                        tokenDB.Expiration = token.Expiration;
                        await _tokenDBService.UpdateTokenAsync(token.Code, tokenDB);
                        CacheMaintenance.Instance.Get<ITokensManager>().Remove(tokenDB.Code);
                    }
                }
                catch (Exception ex)
                {
                    Logger.WriteLineError($"token Update err:{ex}");
                }
            }
            return true;
        }

        private async Task<bool> RemoveExpiredToken()
        {
            if (!ConfigurationManager.IsDistributed || ConfigurationManager.IsMaster)
            {
                //删除db过期token和对应的内存
                var tokens = await _tokenDBService.FindExpiredTokensAsync();
                foreach (var token in tokens)
                {
                    try
                    {
                        await _tokenDBService.RemoveTokenAsync(token.Code);
                        CacheMaintenance.Instance.Get<ITokensManager>().Remove(token.Code);
                    }
                    catch (Exception ex)
                    {
                        Logger.WriteLineError($"RemoveExpiredToken err:{ex}");
                    }
                }
            }
            return true;
        }

        /// <summary>
        /// 获取Vod权限验证
        /// </summary>
        /// <param name="request">验签平台类型</param>
        /// <returns>验证后生成的验签信息</returns>
        public async Task<string> GetVodAuthorizationAsync(TokenRequest request)
        {
            try
            {
                return await Task.Run(() =>
                {
                    var sign = SignatureHelper.GetVodAuthorization();
                    return sign;
                });
            }
            catch (RpcException)
            {
                throw;
            }
            catch (Exception ex)
            {
                ThrowCustomerException(CustomerRpcCode.ExceptionError, ex.Message);
                return string.Empty;
            }
        }

        /// <summary>
        /// 获取权限验证
        /// </summary>
        /// <param name="request">验签平台类型</param>
        /// <returns>验证后生成的验签信息</returns>
        public async Task<string> GetAuthorizationAsync(AuthenticationRequest request)
        {
            try
            {
                if (string.IsNullOrEmpty(request.FileName))
                {
                    ThrowCustomerException(CustomerRpcCode.FileNameIsEmpty, "FileName Empty error");
                }
                return await Task.Run(() =>
                {
                    List<KeyValuePair<string, string>> requestParams = null;
                    List<KeyValuePair<string, string>> headers = null;
                    if (request.UrlParams?.Count > 0)
                    {
                        requestParams = new List<KeyValuePair<string, string>>();
                        foreach (var urlItem in request.UrlParams)
                        {
                            requestParams.Add(
                                new KeyValuePair<string, string>(urlItem.Key, urlItem.Value)
                            );
                        }
                    }
                    if (request.HeaderParams?.Count > 0)
                    {
                        headers = new List<KeyValuePair<string, string>>();
                        foreach (var headItem in request.HeaderParams)
                        {
                            headers.Add(
                                new KeyValuePair<string, string>(headItem.Key, headItem.Value)
                            );
                        }
                    }
                    if (string.IsNullOrEmpty(request.RequestMethod))
                    {
                        request.RequestMethod = "put";
                    }
                    else
                    {
                        request.RequestMethod = request.RequestMethod.ToLower();
                    }
                    var sign = SignatureHelper.GetStorageAuthorization(request.FileName, request.ServerType, request.RequestMethod, requestParams, headers, request.IsCompleted);
                    return sign;
                });
            }
            catch (RpcException)
            {
                throw;
            }
            catch (Exception ex)
            {
                ThrowCustomerException(CustomerRpcCode.ExceptionError, ex.Message);
                return string.Empty;
            }
        }

        /// <summary>
        /// Apply a token for clientw
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<TokenDTO> ApplyTokenAsync(ApplyTokenRequest request)
        {
            //TODO save account name
            var newToken = await _serverTokenManager.AddAsync(request.AccountType, request.AccountName, request.ClientId, request.Platform, request.LoginSource, request.LoginServer, request.IpValue, request.InstallVersion, request.IsOldPlatform, request.ProxyType);
            var newTokenDTO = newToken.MappingTo<TokenDTO>();
            return await Task.FromResult(newTokenDTO);
        }

        /// <summary>
        /// 新增token for db
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="tokenAddRequest"></param>
        /// <returns></returns>
        private async Task<bool> OnTokenAdd(TokenAddRequest tokenAddRequest)
        {
            var newToken = tokenAddRequest.Token;
            if (newToken != null)
            {
                var loginSources = GetLoginSource(tokenAddRequest.LoginSource);
                //save to database
                var oldDBTokens = await _tokenDBService.FindTokensByClentIdAsync(tokenAddRequest.ClientId);
                oldDBTokens = oldDBTokens.Where(x => loginSources.Contains(x.LoginSource))?.ToList();
                if (oldDBTokens != null && oldDBTokens.Count() > 0)
                {
                    foreach (var item in oldDBTokens)
                    {
                        await _tokenDBService.RemoveTokenAsync(item.Code);
                    }
                }
                var newTokenDTO = newToken.MappingTo<TokenDTO>();
                await _tokenDBService.AddTokenInfoAsync(new AddTokenInfoRequest { Data = newTokenDTO });
            }
            return true;
        }

        private List<LoginSource> GetLoginSource(LoginSource loginSource)
        {
            var list = new List<LoginSource>();
            switch (loginSource)
            {
                case LoginSource.PC:
                case LoginSource.Web:
                    list.Add(LoginSource.PC);
                    list.Add(LoginSource.Web);
                    break;
                case LoginSource.Mobile:
                case LoginSource.Pad:
                    list.Add(LoginSource.Mobile);
                    list.Add(LoginSource.Pad);
                    break;
                default:
                    list.Add(loginSource);
                    break;
            }
            return list;
        }

        private List<int> GetLoginSourceInt(LoginSource loginSource)
        {
            var list = new List<int>();
            switch (loginSource)
            {
                case LoginSource.PC:
                case LoginSource.Web:
                    list.Add((int)LoginSource.PC);
                    list.Add((int)LoginSource.Web);
                    break;
                case LoginSource.Mobile:
                case LoginSource.Pad:
                    list.Add((int)LoginSource.Mobile);
                    list.Add((int)LoginSource.Pad);
                    break;
                default:
                    list.Add((int)loginSource);
                    break;
            }
            return list;
        }

        /// <summary>
        /// Validate the token from client
        /// </summary>
        /// <param name="request">Token string</param>
        /// <returns></returns>
        public async Task<ValidateTokenResult> ValidateTokenAsync(ValidateTokenRequest request)
        {
            if (string.IsNullOrWhiteSpace(request.Token))
            {
                return new ValidateTokenResult(CustomerRpcCode.TokenNotExist, new TokenDTO()); //TODO this should be empty token not TokenNotExist
            }
            if (request.Token == DefaultToken)
            {
                return new ValidateTokenResult(CustomerRpcCode.Ok, new TokenDTO());//内部调用，没有client返回
            }
            var result = _serverTokenManager.Validate(request.Token, request.MethodName);
            return await Task.FromResult(result);
        }

        /// <summary>
        /// get tokendto
        /// </summary>
        /// <param name="request">Token string</param>
        /// <returns></returns>
        public async Task<TokenDTO> GetTokenAsync(TokenRequest request)
        {
            if (request.Token == DefaultToken)
            {
                return new TokenDTO();//内部调用，没有client返回
            }
            var result = _serverTokenManager.GetTokenDTO(request.Token);
            return await Task.FromResult(result);
        }

        /// <summary>
        /// Get tokens with specified client Id
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<IList<TokenDTO>> GetTokensWithClientIdAsync(GetTokensWithClientIdRequest request)
        {
            var tokens = _serverTokenManager.GetTokens(request.ClientId);
            return await Task.FromResult(tokens.MappingTo<IList<TokenDTO>>());
        }

        /// <summary>
        /// Get tokens with specified client ids
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<IList<TokenDTO>> GetTokenWithClientIdsAsync(GetTokenWithClientIdsRequest request)
        {
            var tokens = _serverTokenManager.GetTokenWithClientIds(request.ClientIds);
            return await Task.FromResult(tokens.MappingTo<IList<TokenDTO>>());
        }

        /// <summary>
        /// Get tokens with specified token values
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<IList<TokenDTO>> GetTokenWithValuesAsync(GetTokenWithValuesRequest request)
        {
            var tokens = _serverTokenManager.GetTokenWithValues(request.TokenValues);
            return await Task.FromResult(tokens.MappingTo<IList<TokenDTO>>());
        }

        /// <summary>
        /// Set token online state
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SetOnlineStateAsync(SetOnlineStateRequest request)
        {
            var token = request.Token;
            var cacheToken = CacheMaintenance.Instance.Get<ITokensManager>().Get(token);
            if (cacheToken == null || cacheToken.IsOnline != request.IsOnline)
            {
                var tokenInfo = await _tokenDBService.FindTokenByValueAsync(token);
                if (tokenInfo != null)
                {
                    tokenInfo.IsOnline = request.IsOnline;
                    var result = await _tokenDBService.UpdateTokenAsync(tokenInfo.Code, tokenInfo);
                    CacheMaintenance.Instance.Get<ITokensManager>().Remove(token);
                    return result;
                }
                return false;
            }
            return true;
        }

        /// <summary>
        /// Set token ip address
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SetIPAddressAsync(SetIPAddressRequest request)
        {
            var token = request.Token;
            var ipAddress = request.IPAddress;
            var tokenInfo = await _tokenDBService.FindTokenByValueAsync(token);
            if (tokenInfo != null && tokenInfo.IpValue == 0)
            {
                tokenInfo.IpValue = IpToLong(ipAddress);
                var result = await _tokenDBService.UpdateTokenAsync(tokenInfo.Code, tokenInfo);
                CacheMaintenance.Instance.Get<ITokensManager>().Remove(token);
                return result;
            }
            return true;
        }

        /// <summary>
        /// Set token install version address
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> SetTokenInstallVersionAsync(SetTokenInstallVersionRequest request)
        {
            var token = request.Token;
            var tokenInfo = await _tokenDBService.FindTokenByValueAsync(token);
            if (tokenInfo != null)
            {
                tokenInfo.InstallVersion = request.InstallVersion;
                var result = await _tokenDBService.UpdateTokenAsync(tokenInfo.Code, tokenInfo);
                CacheMaintenance.Instance.Get<ITokensManager>().Remove(token);
                return result;
            }
            return true;
        }

        /// <summary>
        /// Ip转Long
        /// </summary>
        /// <param name="ip"></param>
        /// <returns></returns>
        private long IpToLong(string ip)
        {
            try
            {
                if (System.Net.IPAddress.TryParse(ip, out IPAddress ipAddress))
                {
                    char[] separator = new char[] { '.' };
                    string[] items = ip.Split(separator);
                    var ipLong = long.Parse(items[0]) << 24 | long.Parse(items[1]) << 16 | long.Parse(items[2]) << 8 | long.Parse(items[3]);
                    return ipLong;
                }
                // else
                // {
                //     Logger.WriteLineWarn($"IpToLong failed, err ip formatter, ip: {ip}");
                // }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"IpToLong failed, ip: {ip}, ex:{ex}");
            }
            return 0;
        }

        /// <summary>
        /// Log off from token manager
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> LogOffAsync(TokenRequest request)
        {
            var token = request.Token;
            //remove from dtabase
            await _tokenDBService.RemoveTokenAsync(token);
            CacheMaintenance.Instance.Get<ITokensManager>().Remove(token);
            _serverTokenManager.RemoveExtendToken(token);
            return true;
        }

        /// <summary>
        /// Get paged online accounts（Internal Api used on server）
        /// </summary>
        /// <param name="request">The request parameters</param>
        /// <returns>The online account list</returns>
        /// <show>false</show>
        public async Task<PageResult<TokenDTO>> GetPagedTokensAsync(GetPagedTokensOnlineRequest request)
        {
            var pageSize = request.PageSize;
            var pageIndex = request.PageIndex;
            var totalCount = 0;
            var result = _serverTokenManager.GetPagedTokens(request.Keyword, pageIndex, pageSize, request.AccountType, request.LoginSource, request.IsOnline, out totalCount);
            var list = result.MappingTo<IList<TokenDTO>>() as List<TokenDTO>;
            var pageResult = new PageResult<TokenDTO>()
            {
                PageIndex = pageIndex, //why we need this result,oh my god
                PageSize = pageSize, //why
                PageData = list,
                TotalCount = totalCount
            };
            return await Task.FromResult(pageResult);
        }

        public async Task<bool> RefreshBlackInfosAsync(TokenRequest request)
        {
            _serverTokenManager.RefreshBlackClientIds();
            return await Task.FromResult(true);
        }

        /// <summary>
        /// Release manager
        /// </summary>
        protected override void Disposing()
        {
            _serverTokenManager.TokenReplaced -= OnTokenReplaced;
        }

        private static string GetStringParams(string node, string item)
        {
            try
            {
                return EnvironmentConfigManager.GetParammeter<StringParameter>(node, item).Value;
            }
            catch
            {
                return "";
            }
        }
    }
}
